from __future__ import annotations

from typing import Any, List
from urllib.parse import urlencode

from .._http import HttpClient
from .._types import Grant, ListGrantsParams, ListGrantsResponse, VerifiedGrant, DelegateParams
from .._verify import _map_online_verify_to_verified_grant


class GrantsClient:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def get(self, grant_id: str) -> Grant:
        data = self._http.get(f"/v1/grants/{grant_id}")
        return Grant.from_dict(data)

    def list(self, params: ListGrantsParams | None = None) -> ListGrantsResponse:
        qs = _build_query(params.to_dict() if params else {})
        path = f"/v1/grants?{qs}" if qs else "/v1/grants"
        data = self._http.get(path)
        return ListGrantsResponse.from_dict(data)

    def revoke(self, grant_id: str) -> None:
        self._http.delete(f"/v1/grants/{grant_id}")

    def delegate(
        self,
        *,
        parent_grant_token: str,
        sub_agent_id: str,
        scopes: List[str],
        expires_in: str | None = None,
    ) -> Any:
        params = DelegateParams(
            parent_grant_token=parent_grant_token,
            sub_agent_id=sub_agent_id,
            scopes=scopes,
            expires_in=expires_in,
        )
        return self._http.post("/v1/grants/delegate", params.to_dict())

    def verify(self, token: str) -> VerifiedGrant:
        response = self._http.post("/v1/grants/verify", {"token": token})
        raw_token: str = response.get("token", token) if isinstance(response, dict) else token
        return _map_online_verify_to_verified_grant(raw_token)


def _build_query(params: dict[str, object]) -> str:
    filtered = {k: v for k, v in params.items() if v is not None}
    if not filtered:
        return ""
    return urlencode(filtered)



"""Tests for rate limit header parsing."""
from __future__ import annotations

import pytest
import respx
import httpx

from grantex import Grantex, GrantexApiError, RateLimit


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


@respx.mock
def test_last_rate_limit_populated_on_success(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/agents").mock(
        return_value=httpx.Response(
            200,
            json={"agents": [], "total": 0, "page": 1, "pageSize": 20},
            headers={
                "x-ratelimit-limit": "100",
                "x-ratelimit-remaining": "97",
                "x-ratelimit-reset": "1709337600",
            },
        )
    )

    client.agents.list()

    assert client.last_rate_limit is not None
    assert client.last_rate_limit == RateLimit(
        limit=100, remaining=97, reset=1709337600
    )


@respx.mock
def test_rate_limit_on_429_with_retry_after(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/agents").mock(
        return_value=httpx.Response(
            429,
            json={"message": "Rate limit exceeded"},
            headers={
                "x-ratelimit-limit": "20",
                "x-ratelimit-remaining": "0",
                "x-ratelimit-reset": "1709337600",
                "retry-after": "42",
            },
        )
    )

    with pytest.raises(GrantexApiError) as exc_info:
        client.agents.list()

    err = exc_info.value
    assert err.status_code == 429
    assert err.rate_limit is not None
    assert err.rate_limit == RateLimit(
        limit=20, remaining=0, reset=1709337600, retry_after=42
    )


@respx.mock
def test_last_rate_limit_none_when_headers_missing(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/agents").mock(
        return_value=httpx.Response(
            200,
            json={"agents": [], "total": 0, "page": 1, "pageSize": 20},
        )
    )

    client.agents.list()

    assert client.last_rate_limit is None



"""Tests for the VaultClient resource."""
from __future__ import annotations

import json

import httpx
import respx

from grantex import (
    Grantex,
    StoreCredentialParams,
    ExchangeCredentialParams,
    ListVaultCredentialsParams,
)

BASE_URL = "https://api.grantex.dev"

MOCK_CREDENTIAL = {
    "id": "vault_01",
    "principalId": "user_123",
    "service": "google",
    "credentialType": "oauth2",
    "tokenExpiresAt": "2026-04-01T00:00:00Z",
    "metadata": {"email": "test@example.com"},
    "createdAt": "2026-03-01T00:00:00Z",
    "updatedAt": "2026-03-01T00:00:00Z",
}


@respx.mock
def test_store_credential() -> None:
    respx.post(f"{BASE_URL}/v1/vault/credentials").mock(
        return_value=httpx.Response(201, json={
            "id": "vault_01",
            "principalId": "user_123",
            "service": "google",
            "credentialType": "oauth2",
            "createdAt": "2026-03-01T00:00:00Z",
        })
    )

    client = Grantex(api_key="test_key", base_url=BASE_URL)
    result = client.vault.store(StoreCredentialParams(
        principal_id="user_123",
        service="google",
        access_token="ya29.access_token",
        refresh_token="rt_refresh",
    ))

    assert result.id == "vault_01"
    assert result.service == "google"
    assert result.credential_type == "oauth2"

    request = respx.calls.last.request
    body = json.loads(request.content)
    assert body["principalId"] == "user_123"
    assert body["accessToken"] == "ya29.access_token"


@respx.mock
def test_list_credentials_with_filters() -> None:
    respx.get(f"{BASE_URL}/v1/vault/credentials?principalId=user_123&service=google").mock(
        return_value=httpx.Response(200, json={"credentials": [MOCK_CREDENTIAL]})
    )

    client = Grantex(api_key="test_key", base_url=BASE_URL)
    result = client.vault.list(ListVaultCredentialsParams(
        principal_id="user_123",
        service="google",
    ))

    assert len(result.credentials) == 1
    assert result.credentials[0].service == "google"


@respx.mock
def test_get_credential() -> None:
    respx.get(f"{BASE_URL}/v1/vault/credentials/vault_01").mock(
        return_value=httpx.Response(200, json=MOCK_CREDENTIAL)
    )

    client = Grantex(api_key="test_key", base_url=BASE_URL)
    result = client.vault.get("vault_01")

    assert result.id == "vault_01"
    assert result.principal_id == "user_123"


@respx.mock
def test_delete_credential() -> None:
    respx.delete(f"{BASE_URL}/v1/vault/credentials/vault_01").mock(
        return_value=httpx.Response(204)
    )

    client = Grantex(api_key="test_key", base_url=BASE_URL)
    client.vault.delete("vault_01")


@respx.mock
def test_exchange_credential() -> None:
    respx.post(f"{BASE_URL}/v1/vault/credentials/exchange").mock(
        return_value=httpx.Response(200, json={
            "accessToken": "ya29.real_token",
            "service": "google",
            "credentialType": "oauth2",
            "tokenExpiresAt": "2026-04-01T00:00:00Z",
            "metadata": {},
        })
    )

    client = Grantex(api_key="test_key", base_url=BASE_URL)
    result = client.vault.exchange(
        "grant.jwt.token",
        ExchangeCredentialParams(service="google"),
    )

    assert result.access_token == "ya29.real_token"
    assert result.service == "google"

    request = respx.calls.last.request
    assert request.headers["authorization"] == "Bearer grant.jwt.token"
    body = json.loads(request.content)
    assert body["service"] == "google"


@respx.mock
def test_list_credentials_without_filters() -> None:
    respx.get(f"{BASE_URL}/v1/vault/credentials").mock(
        return_value=httpx.Response(200, json={"credentials": []})
    )

    client = Grantex(api_key="test_key", base_url=BASE_URL)
    result = client.vault.list()

    assert len(result.credentials) == 0



from __future__ import annotations

from .._http import HttpClient
from .._types import CreatePrincipalSessionParams, PrincipalSessionResponse


class PrincipalSessionsClient:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def create(self, params: CreatePrincipalSessionParams) -> PrincipalSessionResponse:
        data = self._http.post("/v1/principal-sessions", params.to_dict())
        return PrincipalSessionResponse.from_dict(data)



from __future__ import annotations

from .._http import HttpClient
from .._types import ExchangeTokenParams, ExchangeTokenResponse, RefreshTokenParams, VerifyTokenResponse


class TokensClient:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def exchange(self, params: ExchangeTokenParams) -> ExchangeTokenResponse:
        data = self._http.post("/v1/token", params.to_dict())
        return ExchangeTokenResponse.from_dict(data)

    def refresh(self, params: RefreshTokenParams) -> ExchangeTokenResponse:
        data = self._http.post("/v1/token/refresh", params.to_dict())
        return ExchangeTokenResponse.from_dict(data)

    def verify(self, token: str) -> VerifyTokenResponse:
        data = self._http.post("/v1/tokens/verify", {"token": token})
        return VerifyTokenResponse.from_dict(data)

    def revoke(self, token_id: str) -> None:
        self._http.post("/v1/tokens/revoke", {"jti": token_id})
        return None



from __future__ import annotations

import hashlib
import base64
import os
from dataclasses import dataclass


@dataclass(frozen=True)
class PkceChallenge:
    code_verifier: str
    code_challenge: str
    code_challenge_method: str


def generate_pkce() -> PkceChallenge:
    """Generate a PKCE code verifier and S256 challenge pair.

    Use ``code_challenge`` + ``code_challenge_method`` in the authorize request,
    and ``code_verifier`` in the token exchange request.
    """
    verifier_bytes = os.urandom(32)
    code_verifier = base64.urlsafe_b64encode(verifier_bytes).rstrip(b"=").decode("ascii")
    digest = hashlib.sha256(code_verifier.encode("ascii")).digest()
    code_challenge = base64.urlsafe_b64encode(digest).rstrip(b"=").decode("ascii")
    return PkceChallenge(
        code_verifier=code_verifier,
        code_challenge=code_challenge,
        code_challenge_method="S256",
    )



from ._agents import AgentsClient
from ._audit import AuditClient
from ._grants import GrantsClient
from ._tokens import TokensClient

__all__ = ["AgentsClient", "AuditClient", "GrantsClient", "TokensClient"]



"""Tests for PrincipalSessionsClient."""
from __future__ import annotations

import json

import pytest
import respx
import httpx

from grantex import Grantex, CreatePrincipalSessionParams


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


@respx.mock
def test_create_principal_session(client: Grantex) -> None:
    payload = {
        "sessionToken": "eyJ...",
        "dashboardUrl": "https://api.grantex.dev/permissions?session=eyJ...",
        "expiresAt": "2026-03-01T00:00:00Z",
    }
    route = respx.post("https://api.grantex.dev/v1/principal-sessions").mock(
        return_value=httpx.Response(201, json=payload)
    )

    response = client.principal_sessions.create(
        CreatePrincipalSessionParams(principal_id="user_123", expires_in="2h")
    )

    assert response.session_token == "eyJ..."
    assert response.dashboard_url.endswith("/permissions?session=eyJ...")
    assert response.expires_at == "2026-03-01T00:00:00Z"

    body = json.loads(route.calls[0].request.content)
    assert body["principalId"] == "user_123"
    assert body["expiresIn"] == "2h"


@respx.mock
def test_create_principal_session_minimal(client: Grantex) -> None:
    payload = {
        "sessionToken": "tok",
        "dashboardUrl": "url",
        "expiresAt": "2026-03-01T00:00:00Z",
    }
    route = respx.post("https://api.grantex.dev/v1/principal-sessions").mock(
        return_value=httpx.Response(201, json=payload)
    )

    response = client.principal_sessions.create(
        CreatePrincipalSessionParams(principal_id="user_456")
    )

    assert response.session_token == "tok"

    body = json.loads(route.calls[0].request.content)
    assert body["principalId"] == "user_456"
    assert "expiresIn" not in body
